#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : utils
# @Time         : 2020/4/13 6:49 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


from enum import Enum


# from collections import namedtuple
# FeatType = namedtuple('FeatType', ['Cat', 'Num'])('Cat', 'Num')
class FeatureType(str, Enum):
    DenseFeature = "DenseFeature"
    SparseFeature = "SparseFeature"
    SequenceFeature = "SequenceFeature"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : __init__.py
# @Time         : 2019-10-21 13:00
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

from .Dataset import Dataset


def load_iris(batch_size=128):
    from sklearn import datasets
    ds = Dataset(batch_size)
    X, y = datasets.load_iris(True)
    return ds.from_cache(X, y)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : Add
# @Time         : 2020/4/10 11:20 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 
import tensorflow as tf


# todo: 验证层结构

class Add(tf.keras.layers.Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def build(self, input_shape):
        # Be sure to call this somewhere!
        super().build(input_shape)

    def call(self, inputs, **kwargs):
        if not isinstance(inputs, list):
            return inputs
        if len(inputs) == 1:
            return inputs[0]
        if len(inputs) == 0:
            return tf.constant([[0.0]])

        return tf.keras.layers.add(inputs)


class OurDense(tf.keras.layers.Layer):
    """原来是继承Layer类，现在继承OurLayer类
    """

    def __init__(self, hidden_dim, output_dim,
                 hidden_activation='linear',
                 output_activation='linear', **kwargs):
        super(OurDense, self).__init__(**kwargs)
        self.hidden_dim = hidden_dim
        self.output_dim = output_dim
        self.hidden_activation = hidden_activation
        self.output_activation = output_activation

    def build(self, input_shape):
        """在build方法里边添加需要重用的层，
        当然也可以像标准写法一样条件可训练的权重。
        """
        super(OurDense, self).build(input_shape)
        self.h_dense = tf.keras.layers.Dense(self.hidden_dim,
                                             activation=self.hidden_activation)
        self.o_dense = tf.keras.layers.Dense(self.output_dim,
                                             activation=self.output_activation)

    def call(self, inputs, **kwargs):
        """直接reuse一下层，等价于o_dense(h_dense(inputs))
        :param **kwargs:
        """
        # self.o_dense(self.h_dense(inputs))
        h = self.reuse(self.h_dense, inputs)
        o = self.reuse(self.o_dense, h)
        return o

    def compute_output_shape(self, input_shape):
        return input_shape[:-1] + (self.output_dim,)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : DataLoader
# @Time         : 2019-10-21 13:01
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :
"""
cache
"""
import numpy as np
import pandas as pd
import tensorflow as tf

from pathlib import Path
from functools import partial


# TODO:
#  ds.repeat()?
#
class Dataset(object):
    """
    https://tensorflow.google.cn/guide/data?hl=zh_cn

    """

    def __init__(self, batchsize=128, cache_filename=""):
        self.batchsize = batchsize
        self.cache_filename = cache_filename

    def from_cache(self, X, y=None, shuffle_buffer_size=10000):
        """
        from sklearn.datasets import load_iris
        X, y = load_iris(True)

        :param X: (list, np.ndarray, dict, pd.DataFrame)
        :param y:
        :param shuffle_buffer_size:
        :return:
        """
        assert isinstance(X, (list, np.ndarray, dict, pd.DataFrame)), "Date Type Error"

        if isinstance(X, pd.DataFrame):
            X = X.to_dict('list')
        if y is None:
            tensors = X
        else:
            tensors = (X, y)

        # Common
        ds = tf.data.Dataset.from_tensor_slices(tensors)
        ds = ds.shuffle(shuffle_buffer_size, seed=None)
        ds = ds.batch(self.batchsize)
        ds = ds.prefetch(tf.data.experimental.AUTOTUNE)

        return ds


    def from_generator(self):
        # TODO: 增加对文件的操作（txt/tfrecord）
        # tf.data.Dataset.from_generator()
        pass

    def _from_np_array(self, array):
        # Common
        buffer_size = len(array)
        ds = tf.data.Dataset.from_tensor_slices(array)
        ds = ds.shuffle(buffer_size, seed=None)
        ds = ds.batch(self.batchsize)
        ds = ds.prefetch(tf.data.experimental.AUTOTUNE)

        return ds

    def _from_pd_dataframe(self, df: pd.DataFrame, label="label"):
        """
        import tensorflow as tf
        dataset = tf.data.Dataset.from_tensor_slices(({"a": [1, 2], "b": [3, 4]}, [0, 1]))
        print(list(dataset.as_numpy_iterator()))

        :param df:
        :param label:
        :return:
        """
        if label and label in df.columns:
            df = df.drop(labels=[label], axis=1)
            labels = df[label]
            tensors = (df.to_dict('list'), labels)  # df.to_dict('series')
        else:
            tensors = df.to_dict('list')

        # Common
        buffer_size = len(df)
        ds = tf.data.Dataset.from_tensor_slices(tensors)
        ds = ds.shuffle(buffer_size, seed=None)
        ds = ds.batch(self.batchsize)
        ds = ds.prefetch(tf.data.experimental.AUTOTUNE)

        # features_dataset = tf.data.Dataset.from_tensor_slices(features)
        # labels_dataset = tf.data.Dataset.from_tensor_slices(labels)
        # tf.data.Dataset.zip((features_dataset, labels_dataset))
        return ds


    def from_tfrecord(self, feature_dict: dict, file_pattern: str, file_shuffle=True, file_shuffle_seed=666,
                      shuffle_buffer_size=10000):
        """
        ds = Dataset()
        feature_dict = {
            'id': tf.io.FixedLenFeature((), tf.int64),
            'feature': tf.io.FixedLenFeature((), tf.int64)
            }
        ds = ds.from_tfrecord(feature_dict, '/Users/yuanjie/Desktop/Projects/Spark/MIPush/test-output.tfrecord/part*')

        :param feature_dict:
        :param file_pattern:
        :param shuffle:
        :param seed:
        :param shuffle_buffer_size:
        :return:
        """
        # TODO: cache
        # assert isinstance(filename, str), f"file path error: {filename}"
        #
        # if Path(filename).is_dir():
        #     filename = list(map(str, Path(filename).glob(glob_regex))) # tf.data.Dataset.list_files

        parser_fn = partial(tf.io.parse_single_example, features=feature_dict)
        filenames = tf.data.Dataset.list_files(file_pattern, file_shuffle, file_shuffle_seed)

        ds = (
            tf.data.TFRecordDataset(filenames, num_parallel_reads=tf.data.experimental.AUTOTUNE)
                .map(parser_fn)
                .shuffle(shuffle_buffer_size)
                .batch(self.batchsize)
                .prefetch(buffer_size=tf.data.experimental.AUTOTUNE)  # 若内存泄露，需手动指定
        )
        return ds

